clear
close all
clc

%% Task 4

% 16bit-Quantizer
w       = 16;
Q       = 2^(-(w-1));

% Generate input signals
f_s     = 44100;
N       = 1024;
f_0    	= 64 * f_s / N;
n_x     = 0 : N-1;
x_lin   = -3*Q : Q/100 : 3*Q;
x_sin   = Q * sin(2*pi*f_0/f_s*n_x);

% Truncation
x_lin_qt    = Q * floor(x_lin/Q); 
e_lin_t     = x_lin - x_lin_qt;
x_sin_qt    = Q * floor(x_sin/Q);
e_sin_t     = x_sin - x_sin_qt;

% Rounding
x_lin_qr    = Q * floor(x_lin/Q+0.5); 
e_lin_r     = x_lin - x_lin_qr;
x_sin_qr    = Q * floor(x_sin/Q+0.5);
e_sin_r     = x_sin - x_sin_qr;

% Show nonlinear transfer characteristic and error signal for -3Q<x(n)<3Q
figure;
% Truncation
subplot(2,2,1)
plot(x_lin,x_lin_qt,'b');
grid on;
xlim([-3*Q 3*Q]); ylim([-3*Q 3*Q]);
xticks(-3*Q:Q:3*Q); xticklabels({'-3Q','-2Q','-Q','0','Q','2Q','3Q'});
yticks(-3*Q:Q:3*Q); yticklabels({'-3Q','-2Q','-Q','0','Q','2Q','3Q'});
title('Truncation');
xlabel('x(n)'); ylabel('x_{Q,t}(n)');
subplot(2,2,3)
plot(x_lin,e_lin_t,'r');
grid on;
xlim([-3*Q 3*Q]); ylim([-Q Q]);
xticks(-3*Q:Q:3*Q); xticklabels({'-3Q','-2Q','-Q','0','Q','2Q','3Q'});
yticks(-Q:Q:Q); yticklabels({'-Q','0','Q'});
xlabel('x(n)'); ylabel('e_{t}(n)');
% Rounding
subplot(2,2,2)
plot(x_lin,x_lin_qr,'b');
grid on;
xlim([-3*Q 3*Q]); ylim([-3*Q 3*Q]);
xticks(-3*Q:Q:3*Q); xticklabels({'-3Q','-2Q','-Q','0','Q','2Q','3Q'});
yticks(-3*Q:Q:3*Q); yticklabels({'-3Q','-2Q','-Q','0','Q','2Q','3Q'});
title('Rounding');
xlabel('x(n)'); ylabel('x_{Q,r}(n)');
subplot(2,2,4)
plot(x_lin,e_lin_r,'r');
grid on;
xlim([-3*Q 3*Q]); ylim([-Q Q]);
xticks(-3*Q:Q:3*Q); xticklabels({'-3Q','-2Q','-Q','0','Q','2Q','3Q'});
yticks(-Q:Q:Q); yticklabels({'-Q','0','Q'});
xlabel('x(n)'); ylabel('e_{r}(n)');

% Consider the sine wave x(n)=Q*sin(2*pi*64/N*nx)
n           = 0:99;
NFFT        = N;
X_sin_qt    = fft(x_sin_qt,NFFT);
X_sin_qr    = fft(x_sin_qr,NFFT);
f_vec       = (0:(NFFT-1)) / NFFT * f_s;

figure;
subplot(2,2,1);
plot(n, x_sin(n+1), 'k'); hold on;
stem(n, x_sin_qt(n+1), '.b'); hold off;
title('Truncation');
xlabel('n'); ylabel('x_{Q}(n)');
axis([0 max(n) -6e-5 +6e-5]); grid on;
subplot(2,2,2);
plot(n, x_sin(n+1), 'k'); hold on;
stem(n, x_sin_qr(n+1), '.'); hold off;
title('Rounding');
xlabel('n'); ylabel('x_{Q}(n)');
axis([0 max(n) -6e-5 +6e-5]); grid on;
subplot(2,2,3);
plot(f_vec,20*log10(abs(X_sin_qt))); grid on; xlim([0 f_s/2]); ylim([-110 0]);
xlabel('f'); ylabel('|X_{Q}(f)| in dB');
subplot(2,2,4);
X_sin_qr(X_sin_qr==0) = 10^-6;
plot(f_vec,20*log10(abs(X_sin_qr))); grid on; xlim([0 f_s/2]); ylim([-110 0]);
xlabel('f'); ylabel('|X_{Q}(f)| in dB');

% Quantization Error and SNR
sigma_Et        = 1/length(e_sin_t) * sum(e_sin_t.^2);      % Error power
sigma_Er        = 1/length(e_sin_r) * sum(e_sin_r.^2);
sigma_X         = 1/length(x_sin) * sum(x_sin.^2);          % Signal power
SNRt            = 10* log10(sigma_X/sigma_Et);
SNRr            = 10* log10(sigma_X/sigma_Er);
