clear
close all
clc

%% Task 3

% 16bit-Quantizer
w       = 16;
Q       = 2^(-(w-1));

% Generate input signals
f_s     = 44100;
N       = 1024;
f_0    	= 440; % 64 * f_s / N;
n_x     = 0 : N-1;
A       = Q;
x       = A * sin(2*pi*f_0/f_s*n_x);
f_vec   = (0:N-1) * f_s/N;

% Noise Shaping Filters
h1      = [0 1];
h2      = [0 2 -1];                 % Changed sign of filter coefficients

% Without noise shaping
x_q0        = Q * floor(x/Q+0.5);
e_vec0      = x_q0 - x;

% First-order noise shaping
e_buff      = zeros(length(h1),1);
x_q1        = zeros(1,N);
for n = 1 : N  
    e_filt          = h1 * e_buff;
    x_shaped        = x(n) - e_filt;
    x_q1(n)         = Q * floor(x_shaped/Q+0.5);
    e_new           = x_q1(n) - x_shaped;
    e_buff(1)       = e_new;
    e_buff          = [0; e_buff(1:end-1)];
end
e_vec1      = x_q1 - x;
clear n e_buff x_shaped e_filt e_new

% Second-order noise shaping
e_buff      = zeros(length(h2),1);
x_q2        = zeros(1,N);
for n = 1 : N
    e_filt          = h2 * e_buff;
    x_shaped        = x(n) - e_filt;
    x_q2(n)         = Q * floor(x_shaped/Q+0.5);
    e_new           = x_q2(n) - x_shaped;
    e_buff(1)       = e_new;  
    e_buff          = [0; e_buff(1:end-1)];
end
e_vec2      = x_q2 - x;
clear n e_buff x_shaped e_filt e_new

% Plot
n           = 0 : 99;
figure;
subplot(2,3,1);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q0(n+1), 'b'); 
plot(n, e_vec0(n+1), 'r'); hold off;
title('Without noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,3,4);
plot(f_vec,20*log10(abs(fft(e_vec0,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
subplot(2,3,2);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q1(n+1), 'b'); 
plot(n, e_vec1(n+1), 'r'); hold off;
title('first-order noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,3,5);
plot(f_vec,20*log10(abs(fft(e_vec1,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
subplot(2,3,3);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q2(n+1), 'b'); 
plot(n, e_vec2(n+1), 'r'); hold off;
title('second-order noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,3,6);
plot(f_vec,20*log10(abs(fft(e_vec2,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
