clear
close all
clc

%% Task 3

% 16bit-Quantizer
w       = 16;
Q       = 2^(-(w-1));

% Generate input signals
f_s     = 44100;
N       = 1024;
f_0    	= 440; % 64 * f_s / N;
n_x     = 0 : N-1;
A       = Q;
x       = A * sin(2*pi*f_0/f_s*n_x);
f_vec   = (0:N-1) * f_s/N;

% Noise Shaping Filters
h3      = [0 1.623 -0.982 0.109];
h5      = [0 2.033 -2.165 1.959 -1.590 0.6149];
h9      = [0 2.412 -3.370 3.937 -4.174 3.353 -2.205 1.281 -0.569 0.0847];

% Without noise shaping
x_q0        = Q * floor(x/Q+0.5);
e_vec0      = x_q0 - x;

% Third-order psychoacoustic noise shaping
e_buff      = zeros(length(h3),1);
x_q3        = zeros(1,N);
for n = 1 : N  
    e_filt          = h3 * e_buff;
    x_shaped        = x(n) - e_filt;
    x_q3(n)         = Q * floor(x_shaped/Q+0.5);
    e_new           = x_q3(n) - x_shaped;
    e_buff(1)       = e_new;
    e_buff          = [0; e_buff(1:end-1)];
end
e_vec3      = x_q3 - x;
clear n e_buff x_shaped e_filt e_new

% Fifth-order psychoacoustic noise shaping
e_buff      = zeros(length(h5),1);
x_q5        = zeros(1,N);
for n = 1 : N
    e_filt          = h5 * e_buff;
    x_shaped        = x(n) - e_filt;
    x_q5(n)         = Q * floor(x_shaped/Q+0.5);
    e_new           = x_q5(n) - x_shaped;
    e_buff(1)       = e_new;  
    e_buff          = [0; e_buff(1:end-1)];
end
e_vec5      = x_q5 - x;
clear n e_buff x_shaped e_filt e_new

% Ninth-order noise shaping
e_buff      = zeros(length(h9),1);
x_q9        = zeros(1,N);
for n = 1 : N
    e_filt          = h9 * e_buff;
    x_shaped        = x(n) - e_filt;
    x_q9(n)         = Q * floor(x_shaped/Q+0.5);
    e_new           = x_q9(n) - x_shaped;
    e_buff(1)       = e_new;  
    e_buff          = [0; e_buff(1:end-1)];
end
e_vec9      = x_q9 - x;
clear n e_buff x_shaped e_filt e_new

% Plot
n           = 0 : 99;
figure;
subplot(2,4,1);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q0(n+1), 'b'); 
plot(n, e_vec0(n+1), 'r'); hold off;
title('Without noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,4,5);
plot(f_vec,20*log10(abs(fft(e_vec0,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
subplot(2,4,2);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q3(n+1), 'b'); 
plot(n, e_vec3(n+1), 'r'); hold off;
title('first-order noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,4,6);
plot(f_vec,20*log10(abs(fft(e_vec3,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
subplot(2,4,3);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q5(n+1), 'b'); 
plot(n, e_vec5(n+1), 'r'); hold off;
title('second-order noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,4,7);
plot(f_vec,20*log10(abs(fft(e_vec5,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
subplot(2,4,4);
plot(n, x(n+1), 'k'); hold on;
plot(n, x_q9(n+1), 'b'); 
plot(n, e_vec9(n+1), 'r'); hold off;
title('second-order noise shaping');
legend('x(n)','x_q(n)','e(n)')
xlabel('n'); ylabel('x(n)');
grid on;
subplot(2,4,8);
plot(f_vec,20*log10(abs(fft(e_vec9,N)))); grid on;
xlim([0 f_s/2]); ylim([-120 -40]);
xlabel('f'); ylabel('|E(f)| in dB');
